/*
 * rssyslog
 * The beginnings of what is to grow into a major syslog-server
 * Copyright (C) 2005-2007 Rune Stre
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Rune Stre <rst@folesvaert.no>
 *  
 */

/*
 * decode_message.c
 * Implementation of functions for decoding messages to printable format
 */

#include "decode_message.h"
#include "syslog_names.h"
#include "rssyslogd.h"

#include <string.h>     /* memset, strnlen */
#include <ctype.h>      /* isspace, isupper, ... */
#include <stdio.h>      /* snprintf */

#include <sys/syslog.h>

#include <sys/types.h>  /* inet_ntop */
#include <sys/socket.h> /* inet_ntop */
#include <arpa/inet.h>  /* inet_ntop */

#include <errno.h>


/* Some handy strings */
#define str_nl       "\n"
#define strl_nl      1

#define str_sep1     ": "
#define strl_sep1    2

#define str_sep2     ":"
#define strl_sep2    1

#define str_sep3     " "
#define strl_sep3    1

#define str_sep4     " - "
#define strl_sep4    3

#define str_no_ts   "<no ts>"
#define strl_no_ts  7

#define str_no_host   "<no host>"
#define strl_no_host  9

#define str_unknown  "???"
#define strl_unknown 3

/*
 * Initializes standard separator fields
 */
int init_message(const struct Config* c, struct Message* m)
{
  if(m == NULL)
    return -1;

  memset(m, 0, sizeof(*m));

  m->parts_used = pNumParts;

  m->parts[pLocalTS].iov_base   = m->date_str;
  m->parts[pLocalTS].iov_len    = DATE_STRLEN;

  m->parts[pLocalSep1].iov_base = str_sep3;
  m->parts[pLocalSep1].iov_len  = strl_sep3;

  m->parts[pLocalHost].iov_base = m->ip_str;
  m->parts[pLocalHost].iov_len  = 0;

  m->parts[pLocalSep2].iov_base = str_sep3;
  m->parts[pLocalSep2].iov_len  = strl_sep3;

  m->parts[pSep1].iov_base = c->decode_full ? str_sep1 : NULL; 
  m->parts[pSep1].iov_len  = c->decode_full ? strl_sep1 : 0; 
  
  m->parts[pSep2].iov_base = c->decode_full ? str_sep1 : NULL; 
  m->parts[pSep2].iov_len  = c->decode_full ? strl_sep1 : 0; 

  m->parts[pSep3].iov_base = str_sep2;
  m->parts[pSep3].iov_len  = strl_sep2;

  m->parts[pSep4].iov_base = str_sep4;
  m->parts[pSep4].iov_len  = strl_sep4;
  
  m->parts[pNL].iov_base = str_nl;
  m->parts[pNL].iov_len  = strl_nl;
  
  return 0;
}

int reset_message(const struct Config* c, struct Message* m)
{
  if(m == NULL)
    return -1;

  m->fac = LOG_NFACILITIES;
  m->sev = 0xff;
  m->src_ip = 0;

  m->parts[pTS].iov_base = c->decode_full ? str_no_ts : NULL; 
  m->parts[pTS].iov_len  = c->decode_full ? strl_no_ts : 0; 
  
  m->parts[pHost].iov_base = c->decode_full ? str_no_host : NULL; 
  m->parts[pHost].iov_len  = c->decode_full ? strl_no_host : 0; 

  m->parts[pFac].iov_base = str_unknown;
  m->parts[pFac].iov_len  = strl_unknown;

  m->parts[pSev].iov_base = str_unknown;
  m->parts[pSev].iov_len  = strl_unknown;

  m->parts[pMsg].iov_base = NULL;
  m->parts[pMsg].iov_len  = 0;

  m->parts[pLocalTS].iov_base  = str_unknown;
  m->parts[pLocalTS].iov_len   = strl_unknown;

  m->parts[pLocalHost].iov_base  = str_unknown;
  m->parts[pLocalHost].iov_len   = strl_unknown;

  return 0;
}


int decode_message(const struct Config* c, struct Message* m)
{
  /* A sample message:
   *  "<0>Aug  7 13:42:09 hname progname message message" 
   */

  if(m == NULL)
    return -1;

  int ix = 0;
  int pri = 0;

  m->src_ip = 0;

  /* 
   * IP-address for log-matching 
   */

  /* ipv4-udp */
  if (m->addr_type == IPv4_UDP) {
    if(m->addr_len)
      m->src_ip = ntohl(m->addr.ipv4_udp.sin_addr.s_addr);
  
    /* Printable IP address */
    if (inet_ntop( AF_INET, &(m->addr.ipv4_udp.sin_addr), 
                   m->ip_str, INET_ADDRSTRLEN ) == NULL)
      warn(errno, "Failed to decode IP address");
    else {
      m->parts[pLocalHost].iov_base = m->ip_str;
#ifdef __USE_GNU
      m->parts[pLocalHost].iov_len  = strnlen(m->ip_str, INET_ADDRSTRLEN);
#else
      m->parts[pLocalHost].iov_len  = strlen(m->ip_str);
#endif /* __USE_GNU */
    }
  }

  
  /* Decode local timestamp */
/*    if (gmtime_r(&(m->ts.tv_sec), &(m->date)) == NULL) */
/*      warn(errno, "gmtime_r() failed"); */
  if (localtime_r(&(m->ts.tv_sec), &(m->date)) == NULL)
    warn(errno, "localtime_r() failed");
  else {
    m->parts[pLocalTS].iov_len = 
      snprintf( m->date_str, DATE_STRLEN+1,
                "%02d.%02d.%04d %02d:%02d:%02d.%03d", 
                m->date.tm_mday, m->date.tm_mon + 1, 
                1900 + m->date.tm_year,
                m->date.tm_hour, m->date.tm_min, m->date.tm_sec,
                (int)m->ts.tv_usec / 1000 );
    if (m->parts[pLocalTS].iov_len > DATE_STRLEN)
      m->parts[pLocalTS].iov_len = DATE_STRLEN;
    m->parts[pLocalTS].iov_base = m->date_str;
  }

    
  /* 
   * Start decoding the message 
   */

  /* Strip off trailing whitespace */
  while ( (m->buflen>0) && isspace(m->buf[m->buflen-1]) )
    m->buflen--;

  /* Check for leading '<' */
  if( (m->buflen < 3) || (m->buf[ix] != '<') ) 
    return decode_fail(m, 0);
  
  ix++;
  
  /* Extract priority */
  while(m->buf[ix] != '>') {
    if( (ix>3) && ((m->buf[ix]<'0') || (m->buf[ix]>'9')) )
      return decode_fail(m, 0);
    pri *= 10;
    pri += m->buf[ix] - '0';
    ix++;
  }
  if(ix == 1)
    return decode_fail(m, 0);
  if((pri<0 ) || (pri>((LOG_NFACILITIES<<3)+7)) )
    return decode_fail(m, 0);

  /* ix is now pointing to the enclosing '>'. Step forward. */
  ix++;

  /* Decode priority */
  m->sev = pri & 0x07;
  m->fac = pri >> 3;
  if(m->fac >= LOG_NFACILITIES) /* Which should never happen */
    m->fac = LOG_NFACILITIES; /* Which signifies an error. */

  decode_fac_name(m->fac, &(m->parts[pFac]));
  decode_sev_name(m->sev, &(m->parts[pSev]));

  if (c->decode_full) {
    /* Extract the timestamp. */
    if((16 + ix > m->buflen) || (m->buf[ix+15] != ' '))
      return decode_fail(m, ix);

    if ( !( isupper(m->buf[ix]) &&
            islower(m->buf[ix+1]) &&
            islower(m->buf[ix+2]) &&
            (m->buf[ix+3] == ' ') &&
            (isdigit(m->buf[ix+4]) || (m->buf[ix+4] == ' ')) &&
            isdigit(m->buf[ix+5]) &&
            (m->buf[ix+6] == ' ') &&
            (isdigit(m->buf[ix+7]) || (m->buf[ix+7] == ' ')) && /*non-rfc?*/
            isdigit(m->buf[ix+8]) &&
            (m->buf[ix+9] == ':') &&
            isdigit(m->buf[ix+10]) &&
            isdigit(m->buf[ix+11]) &&
            (m->buf[ix+12] == ':') &&
            isdigit(m->buf[ix+13]) &&
            isdigit(m->buf[ix+14]) ) )
      return decode_fail(m, ix);
    
    m->parts[pTS].iov_base = m->buf + ix;
    m->parts[pTS].iov_len  = 15;
    ix+=16;
  
    /* The hostname part. It ends with a single space character. */
    m->parts[pHost].iov_base = m->buf + ix;
    m->parts[pHost].iov_len  = 0;
    while ((ix<m->buflen) && (m->buf[ix] != ' ')) {
      ix++;
      m->parts[pHost].iov_len++;
    }

    /* Skip the trailing space */
    ix++;
  }

  /* The message */
  if(ix < m->buflen) {
    m->parts[pMsg].iov_base = m->buf + ix;
    m->parts[pMsg].iov_len  = m->buflen - ix;
  }


  /* That's it! */

  return 0;
}





int decode_fail(struct Message* m, int ix)
{
  m->parts[pMsg].iov_base = m->buf + ix;
  m->parts[pMsg].iov_len  = m->buflen - ix;
  return 0;
}


/*  
 * Function to decode a message and fill inn the various output parts.
 * The received message must reside in m->buf
 */

int decode_fac_name(int fac, struct iovec* dest)
{
  if(dest == NULL)
    return -1;

  if ((dest->iov_base = get_facility_name(fac)) == NULL)
    dest->iov_base = str_unknown;

  dest->iov_len = strlen(dest->iov_base);

  return 0;
}

int decode_sev_name(int sev, struct iovec* dest)
{
  if(dest == NULL)
    return -1;

  if ((dest->iov_base = get_priority_name(sev)) == NULL)
    dest->iov_base = str_unknown;

  dest->iov_len = strlen(dest->iov_base);

  return 0;
}


