/*
 * rssyslog
 * Copyright (C) 2005-2007 Rune Stre
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Rune Stre <rst@folesvaert.no>
 *  
 */

/*
 * init_targets.c
 * Initialize target datastructures
 */

#include "init_targets.h"
#include "rssyslogd.h"
#include "syslog_names.h"

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>


#define FILE_KW       "File"
#define ENDFILE_KW    "EndFile"
#define CLIENTS_KW    "Clients"
#define FACILITIES_KW "Facilities"
#define PRIORITIES_KW "Priorities"

struct Config* config_create(const char* conffile)
{

  enum ParserState { PSNone, PSHasTarget };
  
  struct Config* config = NULL;

  FILE* fp = NULL;

  enum ParserState pstate = PSNone;
  char keyword[MAXLINELEN];
  char arguments[MAXLINELEN];


  char fname[MAXLINELEN];
  char clients[MAXLINELEN];
  char facilities[MAXLINELEN];
  char priorities[MAXLINELEN];

  config = (struct Config*) malloc(sizeof(struct Config));
  if (config == NULL)
    fail(-1, "Failed to allocate Config");

  config->targets = NULL;
  config->decode_full = 0;
  config->msgqueue_slots = 128;

  /* Open the config file */
  if (conffile == NULL)
    conffile = CFGFILE;

  fp = fopen(conffile, "r");
  if (fp == NULL)
    fail(errno, "Failed to open config file");

  fname[0]      = '\0';
  clients[0]    = '\0';
  facilities[0] = '\0';
  priorities[0] = '\0';

  /* Scan through the config file */
  while ( cfgparser_next_line(fp, keyword, MAXLINELEN, 
                              arguments, MAXLINELEN) == 0) {

    /* Parsing dependent on parser state */
    switch(pstate) {
    case PSNone:
      if (strcmp(keyword, FILE_KW) == 0) {
        strncpy(fname, arguments, MAXLINELEN);
        pstate = PSHasTarget;
      } else
        fail(0, "File directice expected in config file");
      
      break;

    case PSHasTarget:
      if (strcmp(keyword, CLIENTS_KW) == 0) {
        if (clients[0] != 0)
          fail(0, "Invalid data in config file");
        strncpy(clients, arguments, MAXLINELEN);
      } else if (strcmp(keyword, FACILITIES_KW) == 0) {
        if (facilities[0] != 0)
          fail(0, "Invalid data in config file");
        strncpy(facilities, arguments, MAXLINELEN);
      } else if (strcmp(keyword, PRIORITIES_KW) == 0) {
        if (priorities[0] != 0)
          fail(0, "Invalid data in config file");
        strncpy(priorities, arguments, MAXLINELEN);
      } else if (strcmp(keyword, ENDFILE_KW) == 0) {
        target_add(&config->targets, fname, clients, facilities, priorities);
        pstate = PSNone;
      }
      break;
    } 

  } /* end while (...) */

  if (pstate != PSNone)
    fail(0, "config file ended prematurely");

  fclose(fp);

  return config;
}


int config_delete(struct Config* c)
{
  if (c == NULL)
    return -1;

  targets_delete(c->targets);
  free(c);

  return 0;
}


int targets_delete(struct Target* targets)
{
  while(targets != NULL) {
    struct Target* t = targets->next;

    while (targets->filters) {
      struct MsgFilter* f = targets->filters->next;
      free(targets->filters);
      targets->filters = f;
    }

    close (targets->fd);

    free (targets);

    targets = t;
  }
  return 0;
}


struct MsgFilter* msgfilter_create( const char* hostdesc, 
                           const char* fac, 
                           const char* sev )
{
  struct MsgFilter* t;
  
  t = (struct MsgFilter*) malloc(sizeof(struct MsgFilter));
  if(t == NULL)
    fail(errno, "msgfilter_create(): Failed to allocate MsgFilter");

  memset(t, 0, sizeof(struct MsgFilter));

  if ((hostdesc) && (strcmp(hostdesc, "*") == 0) )
    t->src_ip = 0;
  else {
    t->src_ip   = cfgparser_ipv4_addr(hostdesc);
    t->src_mask = cfgparser_ipv4_mask(hostdesc);
  }

  if (msgfilter_set_facilities(t, fac) < 0) {
    warn(-1, "Invalid facilities in config file");
    free(t);
    return NULL;
  }

  if (msgfilter_set_priorities(t, sev) < 0) {
    warn(0, "Invalid severities in config file");
    free(t);
    return NULL;
  }

  return t;

}





struct Target* target_get(struct Target** targets, const char* fbase)
{
  struct Target* t;

  /* Scan to see if file is already opened */
  for (t=*targets; t; t=t->next)
    if (strcmp(t->fbase, fbase) == 0)
      return t;

  
  /* No. Create new file object */
  t = (struct Target*) malloc(sizeof(struct Target));
  if (t==NULL)
    return NULL;

  strncpy(t->fbase, fbase, MAXLINELEN);
  t->fbase[MAXLINELEN-1] = '\0';
  t->fname[0] = '\0';
  
  /* Decent defaults */
  t->fd = -1;
  t->access_date = 0;
  t->filters = NULL;
 
  /* We have successfully created a new file object. */
  /* Insert it into the list. */
  t->next = *targets;
  *targets = t;

  return t;

}

int target_add( struct Target** targets, char* fname, char* clients, 
                char* facilities, char* priorities)
{
  int len = 0;
  char client[MAXLINELEN];
  char* cl = clients;
  struct Target* t = NULL;
  struct MsgFilter* f = NULL;

  if ((clients[0] == 0) || (facilities[0] == 0) || (priorities[0] == 0))
    fail(0, "Lacking data in config file");

  /* Get or create a target entry */
  t = target_get(targets, fname);
  if (t == NULL)
    fail(0, "Failed to allocate Target entry");

  /* Insert a message filter for each client specified */
  while ((len = cfgparser_extract_word(cl, client, MAXLINELEN)) > 0) {
    cl += len;
    f = msgfilter_create(client, facilities, priorities);
    if (f == NULL)
      fail(0, "Failed to allocate MsgFilter");
        
    /* Insert the MsgFilter into logfile list */
    f->next = t->filters;
    t->filters = f;
  }
      
  /* Clear all temporary strings */
  fname[0]      = '\0';
  clients[0]    = '\0';
  facilities[0] = '\0';
  priorities[0] = '\0';
  
  return 0;
} 



int cfgparser_wash_string(char* str, int len)
{
  int src = 0, dst = 0;
  int has_ws = 1;
  if ( (str == NULL) || (len <= 0) )
    return -1;


  for (src=0; (src<len) && (str[src]!='\0'); src++) {
    char c = str[src];
    if (isalnum(c) || ispunct(c)) {
      str[dst] = c;
      dst++;
      has_ws = 0;
    } else if (isspace(c) && !has_ws) {
      str[dst] = ' ';
      dst++;
      has_ws = 1;
    } else {
      /* Just ignore whitespace or whatever */
    }
  }

  /* Remove trailing space, if any */
  if (has_ws && dst)
    dst--;
  
  if (dst == len)
    dst--;

  str[dst] = '\0';
  return dst;
}


int cfgparser_extract_word(const char* src, char* dst, int len)
{
  int n;
  
  if ((src==NULL) || (dst==NULL) || (len<0))
    return -1;

  if(len==0)
    return 0;


  for (n=0; (n<len) && (src[n]!='\0') && (src[n]!=' '); n++)
    dst[n] = src[n];

  if (n==len) {
    n--;
    dst[n] = '\0';
    return n;     /* This should be the src terminating '\0'*/
  } else if (src[n] == ' ') {
    dst[n] = '\0';
    n++;          /* Step past the space */
    return n;
  } 

  /* This must be the terminating NULL */
  dst[n] = '\0';
  return n;
}


uint32_t cfgparser_ipv4_addr(const char* hostdesc)
{
  uint32_t* addr_ptr;
  uint32_t addr;
  uint32_t mask = cfgparser_ipv4_mask(hostdesc);
  char hostname[MAXLINELEN];
  struct hostent* host = NULL;
  int i=0;

  /* Extract hostname */
  while ( (i<(MAXLINELEN-1)) && 
          (i<strlen(hostdesc)) && 
          (hostdesc[i] != '/') ) {
    hostname[i] = hostdesc[i];
    i++;
  }
  hostname[i] = '\0';

  host = gethostbyname(hostname);

  if(host == NULL)
    fail(h_errno, "cfgparser_ipv4_addr(): gethostbyname failed");

  if(host->h_addrtype != AF_INET)
    fail(0, "cfgparser_ipv4_addr(): Not an IPv4 address");

  if(host->h_length != 4)
    fail(0, "cfgparser_ipv4_addr(): Several IP-adresses");

  addr_ptr = (uint32_t*)host->h_addr_list[0];
  addr = ntohl(*addr_ptr);
  addr &= mask;
  return addr;

}

uint32_t cfgparser_ipv4_mask(const char* hostdesc)
{
  const char* mask_str = hostdesc;
  int mask_bits = -1;
  uint32_t mask = 0xffffffff;

  while ( (mask_str[0] != '\0') && (mask_str[0] != '/') )
    mask_str++;

  if (mask_str[0] == '/') {
    mask_str++;
    mask_bits = atoi(mask_str);
    if ( (mask_bits<0) || (mask_bits>32) )
      fail (0, "cfgparser_ipv4_addr(): Invalid netmask");
    mask = (mask >> mask_bits) ^ 0xffffffff;
  }

  return mask;

}




int msgfilter_set_facilities(struct MsgFilter* t, const char* s)
{
  char f[MAXLINELEN];
  int offset = 0;
  
  if (t == NULL)
    return -1;
  if (s == NULL)
    return -1;

  if (strcmp(s, "all") == 0) {
    int i;
    for (i=0; i<LOG_NFACILITIES+1; i++)
      t->facilities[i] = 1;
    return 0;
  }

  while ((offset = cfgparser_extract_word(s, f, MAXLINELEN)) > 0) {
    int fac = get_facility_number(f);
    if ( (fac<0) || (fac>LOG_NFACILITIES) )
      return -1;

    t->facilities[fac] = 1;
    s += offset;
  }
    
  return 0;

}



int msgfilter_set_priorities(struct MsgFilter* t, const char* s)
{
  char f[MAXLINELEN];
  int offset = 0;
  
  if (t == NULL)
    return -1;
  if (s == NULL)
    return -1;

  if (strcmp(s, "all") == 0) {
    t->priorities = 0xff;
    return 0;
  }

  while ((offset = cfgparser_extract_word(s, f, MAXLINELEN)) > 0) {
    int sev = get_priority_number(f);
    if ((sev < 0) || (sev > 7))
      return -1;
    t->priorities |= (1 << sev);
    s += offset;
  }
    
  return 0;

}


int cfgparser_next_line( FILE* f, 
                         char* kw, int kwsize, 
                         char* arg, int argsize)
{
  int bufsize = kwsize + argsize;
  char buf[bufsize];

  if (f == NULL)
    return -1;

  while(fgets(buf, bufsize, f)) {
    /* We should check if the line was larger than buf */

    int len = cfgparser_wash_string(buf, bufsize);
    int offset = 0;
    if (len < 0)      
      fail(0, "Invalid data in config file");

    if (len == 0) /* empty line */
      continue;
    
    if (buf[0] == '#') /* comment */
      continue;

    /* We have a valid config line */
    offset = cfgparser_extract_word(buf, kw, kwsize);
    len -= offset;
    if (len >= argsize)
      warn (-1, "config line length exeeds maximum. Trunkating");

    strncpy(arg, buf + offset, argsize);
    arg[len] = 0; /* just in case */
    return 0;
  }

  return -1;
}


