/*
 * rssyslog
 * Copyright (C) 2005-2007 Rune Stre
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Rune Stre <rst@folesvaert.no>
 *  
 */

/*
 * init_targets.h
 * This file is part of the rssyslogd-project, which is intended to
 * grow into a major syslog-server.
 */

#ifndef rssyslog_init_targets_h
#define rssyslog_init_targets_h

/*
 * TODO:
 *
 * DONE:
 *  - change name FileDescriptor -> Target
 *  - change name get_file() -> target_get()
 *  - convert init_logfiles() -> targets_create()
 *  - add functions Target* targets_create() and int targets_delete()
 *
 *  - change name Target -> MsgFilter
 *  - change name new_target() -> msgfilter_create()
 *  - change name set_facilities() -> msgfilter_set_facilities()
 *  - change name set_priorities() -> msgfilter_set_priorities()
 *  - create function int msgfilter_match(const MsgFilter*, Message*)
 *
 *  - change name wash_string() -> cfgparser_wash_string()
 *  - change name extract_word() -> cfgparser_extract_word()
 *  - change name get_ipv4_addr() -> cfgparser_ipv4_addr()
 *  - change name get_ipv4_mask() -> cfgparser_ipv4_mask()
 *
 */


/*  Defines used by the environment (other include-files) */
/*  #define _LARGEFILE64_SOURCE 1 */

#include <stdio.h>
#include <sys/types.h>
#include <stdint.h>
#include <sys/syslog.h>

#define MINI_STRLEN 32
#define MAXLINELEN 1024

struct MsgFilter {
  struct MsgFilter* next;
  uint32_t src_ip;
  uint32_t src_mask;
  u_char facilities[LOG_NFACILITIES+1];
  u_char priorities; /* bitmask */
};

struct Target {
  int fd;
  char fbase[MAXLINELEN];
  char fname[MAXLINELEN];
  uint64_t access_date;
  struct MsgFilter* filters;
  struct Target* next;
};

struct Config {
  struct Target* targets;
  int decode_full;
  int msgqueue_slots;
};

/**************** Functions to operate on Config *********************/

struct Config* config_create(const char* conffile);
int config_delete(struct Config* c);



/**************** Functions to operate on Target *********************/

/*
 * Deletes all targets in chain and associated filters
 */
int targets_delete(struct Target* targets);

/* 
 * Returns a Target for the specified filename template (fbase) 
 * If none is found in the Target list a new one is created and 
 * inserted into the list.
 */
struct Target* target_get(struct Target** targets, const char* fbase);


/*
 * Adds a target and the filters as defined by 
 * clients, facilities and priorities. 
 * Clears strings upon return.
 */
int target_add( struct Target** targets, char* fname, 
                char* clients, 
                char* facilities, 
                char* priorities);


/**************** Message filter functions ****************************/

/* 
 * Creates a new MsgFilter instance with the specified characteristics.
 * Mostly for use by init_logfiles() 
 */
struct MsgFilter* msgfilter_create(const char* hostdesc, 
                                   const char* fac, const char* sev);

/* Extract priorities and facilities from a text string 
 * and set them in a struct MsgFilter.
 * Returns 0 on success, -1 on error */
int msgfilter_set_facilities(struct MsgFilter* t, const char* s);
int msgfilter_set_priorities(struct MsgFilter* t, const char* s);


/**************** config parser helper functions **********************/

/*
 * Removes whitespace from beginning and end of line
 * Replaces multiple whitespace characters with a single space
 * Deletes illegal characters
 * Returns resulting stringlength on success, -1 on error
 */
int cfgparser_wash_string(char* str, int len);

/* 
 * This function expects src to be washed with wash_string 
 * Copies the first word in src to dst, and returns the offset to 
 * the next word in src.
 */
int cfgparser_extract_word(const char* src, char* dst, int len);


/* Calculate IP address and netmask from config file */
uint32_t cfgparser_ipv4_addr(const char* hostdesc);
uint32_t cfgparser_ipv4_mask(const char* hostdesc);

/* 
 * Fetches next non-empty config line and washes it.
 * puts the first word in opt and the rest in arg.
 * Returns 0 on success, -1 on error or end of file.
 */
int cfgparser_next_line( FILE* f, 
                         char* opt, int optsize, 
                         char* arg, int argsize);


#endif /* rssyslog_init_targets_h */



