/*
 * rssyslog
 * The beginnings of what is to grow into a major syslog-server
 * Copyright (C) 2007 Rune Stre
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Rune Stre <rst@folesvaert.no>
 *  
 */

/*
 * logfile_writer.c
 * Implementation of logfile writer
 */


#include "logfile_writer.h"
#include "decode_message.h"
#include "init_targets.h"
#include "rssyslogd.h" /* for warn() and fail() */

#include "rs_thread/rs_thread.h"

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>


int writer_init( struct WriterData* d,
                 struct MsgQueue* q,
                 struct Config* c )
{
  if (d == NULL)
    return -1;

  if (q == NULL)
    return -1;
  
  if (c == NULL)
    return -1;

  d->q = q;
  d->c = c;
  d->msg_counter = 0;
  d->bailout = 0;

  return 0;
}

int target_match_filter(struct Target* t, struct Message* m)
{
  struct MsgFilter* f = NULL;

  if (t == NULL)
    return 0;
    
  for (f=t->filters; f != NULL; f=f->next)
    if (msgfilter_match(f, m))
      return 1;

  return 0;
}



int msgfilter_match(struct MsgFilter* f, struct Message* m)
{
  /* Return 0 if facility doesn't match */
  if (!f->facilities[m->fac])
    return 0;

  /* Return 0 if severity doesn't match */
  if ( !((1<<m->sev) & f->priorities) )
    return 0;

  /* Return 0 if IP address doesn't match */
  if(f->src_ip && ((m->src_ip & f->src_mask) != f->src_ip))
    return 0;

  return 1;
}


int target_get_fd(struct Target* t, struct tm* date)
{
  uint64_t access_date;

  if (t == NULL || date == NULL)
    return -1;

  /* Calculate access date used with logfile rotation */
  access_date = 
    (1900 + date->tm_year) * 100000 + 
    (date->tm_mon + 1) * 100 + 
    date->tm_mday;

  if (t->fd >= 0 && t->access_date == access_date) {
    return t->fd;
  }

  /* File needs to be reopened */
  if (t->fd > 0)
    close (t->fd);

  snprintf(t->fname, MAXLINELEN, "%s_%04d%02d%02d.log", t->fbase, 
           1900 + date->tm_year, date->tm_mon+1, date->tm_mday);
  t->fname[MAXLINELEN-1] = '\0';

  t->fd = open( t->fname, 
                O_WRONLY|O_CREAT|O_APPEND|O_NOCTTY|O_LARGEFILE,
                S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
  t->access_date = access_date;

  if (t->fd<0)
    warn(0, "Failed to open logfile!");

  return t->fd;

}

int write_message(struct Config* conf, struct Message* m)
{

  if(m == NULL)
    return -1;

  if (conf->targets == NULL)
    return -1;

  /* Should we check for properly decoded message? */

  struct Target* t;
  for (t = conf->targets; t; t=t->next) {
    if (target_match_filter(t, m)) {
      /* The message is going into this target. */
      int fd = target_get_fd(t, &m->date);
      if (fd < 0)
        warn(errno, "write_message(): failed to get fd");
      else {
        /* Write the message, try once to reopen if we fail */
        if (writev(t->fd, m->parts, m->parts_used) < 0) {
          warn(errno, "write_message(): writev() failed, "
               "reopening file and retrying...");
          close(t->fd);
          t->fd = -1;
          if ( (fd = target_get_fd(t, &m->date)) < 0)
            warn(errno, "write_message(): failed to reopen file");
          else if (writev(t->fd, m->parts, m->parts_used) < 0) {
            warn(errno, "write_message(): writev() failed. Giving up");
            close(t->fd);
            t->fd = -1;
          }
        } /* end if (writev(...)) */
      } /* end else */
    } /* end if (target_match_filter(...)) */
  } /* end for (...) */
  
  return 0;
}


int writer_thread(void* user_data)
{
  struct WriterData* w = (struct WriterData*) user_data;
  
  if (w == NULL)
    return -1;

  while (!w->bailout) {
    struct Message* m = msgqueue_reader_get(w->q);
    if (m != NULL) {
      if (decode_message(w->c, m) < 0)
        warn(-1, "decode_message() failed");
      else
        write_message(w->c, m);
      
      reset_message(w->c, m);
      msgqueue_reader_release(w->q);
    }
  }

  return 0;
}
