/*
 * rssyslog
 * The beginnings of what is to grow into a major syslog-server
 * Copyright (C) 2007 Rune Stre
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Rune Stre <rst@folesvaert.no>
 *  
 */

/*
 * msg_queue.c
 * Implementation of message queuing.
 */


#include "msg_queue.h"
#include "decode_message.h"

#include <stdlib.h>
#include <sys/time.h>
#include <time.h>
#include <errno.h>


struct MsgQueue* msgqueue_init(const struct Config* c)
{
  int i=0;

  struct MsgQueue* q = malloc(sizeof(struct MsgQueue));
  if (q == NULL)
    return NULL;

  pthread_mutex_init(&q->lock, NULL);
  pthread_cond_init(&q->cond_full, NULL);
  pthread_cond_init(&q->cond_empty, NULL);

  q->n       = c->msgqueue_slots;
  q->wr_head = 0;
  q->wr_tail = 0;
  q->rd_head = 0;
  q->rd_tail = 0;

  q->n_empty = c->msgqueue_slots;
  q->n_full  = 0;
  q->bailout = 0;

  q->q = malloc(sizeof(struct Message)*q->n);
  if (q->q == NULL) {
    free(q);
    return NULL;
  }

  for (i=0; i<q->n; i++)
    init_message(c, &q->q[i]);

  return q;
}


int msgqueue_destroy(struct MsgQueue* q)
{
  /* 
   * Assume noone is using the queue any longer.
   */

  if (q == NULL)
    return -1;

  pthread_cond_destroy(&q->cond_full);
  pthread_cond_destroy(&q->cond_empty);
  pthread_mutex_destroy(&q->lock);
  free(q->q);
  free(q);
  
  return 0;
}



int msgqueue_bailout(struct MsgQueue* q)
{
  if (q == NULL)
    return -1;

  pthread_mutex_lock(&q->lock);
  q->bailout = 1;
  pthread_cond_broadcast(&q->cond_full);
  pthread_cond_broadcast(&q->cond_empty);
  pthread_mutex_unlock(&q->lock);

  return 0;
}


struct Message* msgqueue_writer_get(struct MsgQueue* q)
{
  struct Message* m = NULL;
  struct timespec timeout;
  struct timeval now;
  int bailout = 0;

  if (q == NULL)
    return NULL;

  pthread_mutex_lock(&q->lock);

  /* wait for an empty slot to become available */
  while (q->n_empty == 0 && q->bailout == 0) {
    gettimeofday(&now, NULL);
    timeout.tv_sec = now.tv_sec + 3600; /* wake up once every hour */
    timeout.tv_nsec = now.tv_usec * 1000;
    if (pthread_cond_timedwait(&q->cond_empty, &q->lock, &timeout) == EINTR)
      bailout = 1;
  }


  if (q->n_empty > 0) {
    m = &q->q[q->wr_head];
    q->wr_head++;
    q->wr_head %= q->n;
    q->n_empty--;
  }

  pthread_mutex_unlock(&q->lock);
  return m;
}

struct Message* msgqueue_writer_get_nowait(struct MsgQueue* q)
{
  /* This is supposed to be nonblocking, 
   * but we still have to acquire the mutex lock to maintain integrity. */

  struct Message* m = NULL;

  if (q == NULL)
    return NULL;

  pthread_mutex_lock(&q->lock);

  if (q->n_empty > 0) {
    m = &q->q[q->wr_head];
    q->wr_head++;
    q->wr_head %= q->n;
    q->n_empty--;
  }

  pthread_mutex_unlock(&q->lock);
  return m;

}

int msgqueue_writer_release(struct MsgQueue* q)
{
  if (q == NULL)
    return -1;

  /* This currently does no sanity checking */

  pthread_mutex_lock(&q->lock);

  q->wr_tail = (q->wr_tail+1) % q->n;
  q->n_full++;
  pthread_cond_broadcast(&q->cond_full);

  pthread_mutex_unlock(&q->lock);

  return 0;

}




struct Message* msgqueue_reader_get(struct MsgQueue* q)
{
  struct Message* m = NULL;
  struct timespec timeout;
  struct timeval now;
  int bailout = 0;

  if (q == NULL)
    return NULL;

  pthread_mutex_lock(&q->lock);

  /* wait for an empty slot to become available */
  while (q->n_full == 0 && q->bailout == 0) {
    gettimeofday(&now, NULL);
    timeout.tv_sec = now.tv_sec + 3600; /* wake up once every hour */
    timeout.tv_nsec = now.tv_usec * 1000;
    if (pthread_cond_timedwait(&q->cond_full, &q->lock, &timeout) == EINTR)
      bailout = 1;
  }

  if (q->n_full > 0) {
    m = &q->q[q->rd_head];
    q->rd_head++;
    q->rd_head %= q->n;
    q->n_full--;
  }

  pthread_mutex_unlock(&q->lock);
  return m;
}

struct Message* msgqueue_reader_get_nowait(struct MsgQueue* q)
{
  /* This is supposed to be nonblocking, 
   * but we still have to acquire the mutex lock to maintain integrity. */

  struct Message* m = NULL;

  if (q == NULL)
    return NULL;

  pthread_mutex_lock(&q->lock);

  if (q->n_full > 0) {
    m = &q->q[q->rd_head];
    q->rd_head++;
    q->rd_head %= q->n;
    q->n_full--;
  }

  pthread_mutex_unlock(&q->lock);
  return m;
}


int msgqueue_reader_release(struct MsgQueue* q)
{
  if (q == NULL)
    return -1;

  /* This currently does no sanity checking */

  pthread_mutex_lock(&q->lock);

  q->rd_tail++;
  q->rd_tail %= q->n;
  q->n_empty++;
  pthread_cond_broadcast(&q->cond_empty);

  pthread_mutex_unlock(&q->lock);

  return 0;
}


