/*
 * rssyslog
 * The beginnings of what is to grow into a major syslog-server
 * Copyright (C) 2005-2007 Rune Stre
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Rune Stre <rst@folesvaert.no>
 *  
 */

/*
 * rssyslogd.c
 * The main .c file. Most things are here.
 */

#include "message.h"
#include "receive_message.h"
#include "logfile_writer.h"
#include "decode_message.h"
#include "rssyslogd.h"
#include "init_targets.h"
#include "msg_queue.h"
#include "rs_thread/rs_thread.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>

#include <sys/time.h>
#include <time.h>
#include <sys/ioctl.h>

#include <sys/syslog.h>


/* Local defines */

#ifndef DEV_TTY
#define DEV_TTY  "/dev/tty"
#endif

volatile sig_atomic_t quit_flag = 0;



/* 
 * The TERM signal handler sets a quit flag, 
 * and the program should exit normally
 */
RETSIGTYPE sighandler_term(int sig)
{
    quit_flag = 1;
}

/*
 * These are of course quite useless when the daemon has disconnected from
 * the controlling terminal, and stderr goes nowhere..
 */

void warn(int err, char* str)
{
  if(err > 0)
    fprintf(stderr, "WARNING: %d: %s: %s\n", err, str, strerror(err));
  else
    fprintf(stderr, "WARNING: %d: %s\n", err, str);
}

void fail(int err, char* str)
{
  if(err > 0)
    fprintf(stderr, "Program died: %d: %s: %s\n", err, str, strerror(err));
  else
    fprintf(stderr, "Program died: %d: %s\n", err, str);
  exit (1);
}




int daemonize()
{

  if (fork()) {
    /* parent */
    exit(0);
  } else {

    /* We are the child. Disconnect. */
    int fd = open(DEV_TTY, O_RDWR);
    if (fd < 0) {
      warn(0, "We have seemingly closed the controlling terminal already...");
      return -1;
    } else {
      ioctl(fd, TIOCNOTTY, NULL);
      close(fd);
    }
  }
  return 0;
}




int main(int argc, char* argv[])
{
  struct Config* conf = NULL;      /* The config and targets */
  struct MsgQueue* msg_q = NULL;

  struct ListenerData l_data;
  struct RSThread     l_thr;

  struct WriterData   w_data;
  struct RSThread     w_thr;
  
  /* set up signal handlers */
  signal(SIGTERM, sighandler_term);
  signal(SIGINT,  sighandler_term);
  signal(SIGQUIT, sighandler_term);
  
  /* Read in config */
  conf = config_create(NULL);
  if (conf == NULL)
    fail(-1, "Failed to read config");

  /* Initialize message queue */
  msg_q = msgqueue_init(conf);
  if (msg_q == NULL)
    fail( -1, "Failed to initialize message queue");

  /* Initialize listener */
  if (listener_init( &l_data, msg_q, conf) < 0)
    fail (-1, "Failed to initialize listener data structure");
  if (rs_thread_init (&l_thr, listener_thread, (void*)&l_data) < 0)
    fail (-1, "Failed to initialize listener thread");

  /* Initialize logfile writer */
  if (writer_init(&w_data, msg_q, conf) < 0)
    fail (-1, "Failed to initialize writer data structure");
  if (rs_thread_init (&w_thr, writer_thread, (void*)&w_data) < 0)
    fail (-1, "Failed to initialize writer thread");

  /* Disconnect from the controlling terminal */
  daemonize();

  /* Start the worker threads */
  if (rs_thread_start(&w_thr) < 0)
    fail (errno, "Failed to start writer thread");

  if (rs_thread_start(&l_thr) < 0)
    fail (errno, "Failed to start listener thread");


  /* Wait for a signal to quit has been received */
  while(sleep(5), quit_flag == 0);

  l_data.bailout = 1;
  w_data.bailout = 1;
  msgqueue_bailout(msg_q);

  /* Send a signal to make sure threads are quitting */
  kill(0, SIGTERM);  

  rs_thread_wait(&l_thr);
  rs_thread_wait(&w_thr);

  listener_close(&l_data);
  msgqueue_destroy(msg_q);

  config_delete(conf);

  return 0;
}

